import 'dart:async';
import 'dart:convert' show json, Encoding;
import 'dart:io';

import 'package:connecto_flutter/IdentifyEventModel.dart';
import 'package:device_info/device_info.dart';
import 'package:flutter/foundation.dart';
import 'package:http/http.dart';
import 'package:package_info/package_info.dart';
import 'package:uuid/uuid.dart';

import 'TrackEventModel.dart';

class Connecto {
  static Connecto _singleton;
  static String endpoint = "https://api.connecto.io";
  static String writeKey;
  static String userId;
  static _AnalyticsClient client = new _AnalyticsClient(new Client());
  var uuid = Uuid();

  static String appName;
  static String applicationPackageName;
  String _appVersionName;
  String _appVersionCode;
  String _os;
  String _deviceBrand;
  String _deviceModel;
  String _deviceVersion;
  String _deviceID;
  static bool isDebugEnable = false;

  String _createCryptoRandomString([int length = 32]) {
    return uuid.v4();
  }

//  String getTimeStamp() {
//    return '${DateFormat("yyyy-MM-dd'T'HH:mm:ss:mmm'Z'").format(DateTime.now())}';
//  }

  factory Connecto(
      {@required String writeKey,
      @required String userId,
      @required String applicationName,
      bool isDebugModeEnable,
      @required String packageName}) {
    appName = applicationName;
    applicationPackageName = packageName;
    isDebugEnable = isDebugModeEnable;
    if (_singleton == null) {
      _singleton = new Connecto._internal(writeKey, userId);
      return _singleton;
    } else {
      return _singleton;
    }
  }

  Connecto._internal(String key, String id) {
    writeKey = key;
    userId = id;
//    identify();
    _setBasicInfo();
  }

  void _setBasicInfo() async {
    // set Device Info
    DeviceInfoPlugin deviceInfoPlugin = DeviceInfoPlugin();

    if (Platform.isAndroid) {
      AndroidDeviceInfo androidDeviceInfo = await deviceInfoPlugin.androidInfo;
      _os = "Android";
      _deviceBrand = androidDeviceInfo.brand;
      _deviceModel = androidDeviceInfo.model;
      _deviceVersion = androidDeviceInfo.version.sdkInt.toString();
      _deviceID = androidDeviceInfo.androidId;
    } else {
      IosDeviceInfo iosDeviceInfo = await deviceInfoPlugin.iosInfo;
      _os = "IOS";
      _deviceBrand = iosDeviceInfo.name;
      _deviceModel = iosDeviceInfo.model;
      _deviceVersion = iosDeviceInfo.systemVersion;
      _deviceID = iosDeviceInfo.identifierForVendor;
    }
    // set Package Info

    PackageInfo info = await PackageInfo.fromPlatform();

    _appVersionCode = info.buildNumber;
    _appVersionName = info.version;

//    identify(properties: {"connecto_id": "123456789"});
  }

  void trackEvent(
      {@required String event, Map<String, dynamic> properties = const {}}) {
    Batch batch = new Batch(
        context: _getEventContext(),
        messageId: _createCryptoRandomString(),
        anonymousId: userId,
        channel: "mobile",
        writeKey: writeKey,
        userId: userId,
        properties: properties,
        event: event,
        type: "track",
        timestamp: DateTime.now());
    List<Batch> batchList = new List();
    batchList.add(batch);
    TrackEvent trackEvent = new TrackEvent(
        context: _getEventContext(), sentAt: DateTime.now(), batch: batchList);
    client._postSilentMicroTask("$endpoint/event",
        body: json.encode(trackEvent));
  }

  void trackScreen(
      {@required String screenName,
      Map<String, dynamic> properties = const {}}) {
    Batch batch = new Batch(
        context: _getEventContext(),
        messageId: _createCryptoRandomString(),
        anonymousId: userId,
        channel: "mobile",
        writeKey: writeKey,
        userId: userId,
        properties: properties,
        event: screenName,
        type: "track",
        timestamp: DateTime.now());
    List<Batch> batchList = new List();
    batchList.add(batch);
    TrackEvent trackScreen = new TrackEvent(
        context: _getEventContext(), sentAt: DateTime.now(), batch: batchList);
    client._postSilentMicroTask("$endpoint/event",
        body: json.encode(trackScreen));
  }

  void identify({@required Map<String, dynamic> traits}) {
    IdentifyBatch identifyBatch = new IdentifyBatch(
        traits: traits,
        messageId: _createCryptoRandomString(),
        anonymousId: userId,
        channel: "mobile",
        type: "identify",
        userId: userId,
        writeKey: writeKey,
        timestamp: DateTime.now().toIso8601String(),
        context: _getIdentifyContext(traits));
    List<IdentifyBatch> batchList = new List();
    batchList.add(identifyBatch);
    IdentifyEventModel identifyEventModel =
        new IdentifyEventModel(batch: batchList);

    client._postSilentMicroTask("$endpoint/event",
        body: json.encode(identifyEventModel));
  }

  void trackFcmToken(String token) {
    Map innerValues = {
      "gcm_id": token,
      "app": applicationPackageName + "_" + appName
    };
    Map<String, dynamic> mp = {"\$android_devices": json.encode(innerValues)};
    identify(traits: mp);
  }

  IdentifyContext _getIdentifyContext(Map<String, dynamic> traits) {
    IdentifyContext context = new IdentifyContext(
        app: _getApp(),
        contextLibrary: _getContextLibrary(),
        os: _getContextOs(),
        screen: _getScreenResolution(),
        device: _getDeviceInfo(),
        network: new Network(
            wifi: false, cellular: true, carrier: "Airtel", bluetooth: false),
        writeKey: writeKey,
        traits: traits,
        locale: "en-IN",
        timezone: DateTime.now().toIso8601String(),
        userAgent:
            "Mozilla\/5.0 ($_os; CPU $_os OS $_deviceVersion like Mac OS X) "
            "AppleWebKit\/605.1.15 (KHTML, like Gecko) Mobile\/15E148");
    return context;
  }

  Context _getEventContext() {
    Context context = new Context(
        app: _getApp(),
        contextLibrary: _getContextLibrary(),
        os: _getContextOs(),
        screen: _getScreenResolution(),
        device: _getDeviceInfo(),
        ip: "",
        locale: "en-IN",
        timezone: DateTime.now().toIso8601String(),
        userAgent:
            "Mozilla\/5.0 ($_os; CPU $_os OS $_deviceVersion like Mac OS X) "
            "AppleWebKit\/605.1.15 (KHTML, like Gecko) Mobile\/15E148");
    return context;
  }

  App _getApp() {
    App app = new App(
        name: appName,
        build: _appVersionCode,
        namespace: applicationPackageName,
        version: _appVersionName);
    return app;
  }

  Library _getContextLibrary() {
    Library library =
        new Library(version: _deviceVersion, name: "connecto-flutter");
    return library;
  }

  Screen _getScreenResolution() {
    Screen screen = new Screen(height: 800, width: 400);
    return screen;
  }

  Library _getContextOs() {
    Library library = new Library(version: _deviceVersion, name: "$_os");
    return library;
  }

  Device _getDeviceInfo() {
    Device device = new Device(
      model: _deviceModel,
      id: _deviceID,
      manufacturer: _deviceBrand,
    );
    return device;
  }
}

class _AnalyticsClient extends BaseClient {
  String userAgent;
  Client _inner;

  _AnalyticsClient(this._inner);

  Future<StreamedResponse> send(BaseRequest request) {
    request.headers['Content-Type'] = "application/json";
    request.headers["Authorization"] = Connecto.writeKey;

    if (Connecto.isDebugEnable) print(request.headers);

    return _inner.send(request);
  }

  _postSilentMicroTask(url,
      {Map<String, String> headers, body, Encoding encoding}) async {
    scheduleMicrotask(() async {
      if (Connecto.isDebugEnable) debugPrint("Body : $body", wrapWidth: 1024);
      await this
          .post(url, headers: headers, body: body, encoding: encoding)
          .then((value) {
        if (Connecto.isDebugEnable) {
          print(value.body);
        }
      });
    });
  }
}
